/*
 *
 *	Utility for editting quotas
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <paths.h>

#include <linux/quota.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"
#include "quota_io.h"
#include "quotasys.h"

#define VI_PATH "vi"
#define MAXLINELEN 256

#define FL_COPY 1
#define FL_GRACE 2

int type, flags;
int names;
qid_t protoid;

void usage(char *utilname)
{
	printf("Utility for editting user/group quotas.\nUsage:\n%s [-u | -g] [-p modelname] name...\n", utilname);
	printf("%s -t [-u | -g]\n", utilname);
	printf("Bugs to %s\n", MY_EMAIL);
	exit(1);
}

void parse_options(int argcnt, char **argstr)
{
	int ret;
	char *slash = strrchr(argstr[0], '/');

	if (!slash)
		slash = argstr[0];
	else
		slash++;

	type = QS_USRQUOTA;
	while ((ret = getopt(argcnt, argstr, "Vughtp:")) != EOF) {
		switch (ret) {
			case '?':
			case 'h':
				usage(slash);
			case 'V':
				version();
				exit(0);
			case 'g':
				type = QS_GRPQUOTA;
				break;
			case 'u':
				type = QS_USRQUOTA;
				break;
			case 't':
				flags |= FL_GRACE;
				break;
			case 'p':
				flags |= FL_COPY;
				if (type == QS_USRQUOTA)
					protoid = name2uid(optarg);
				else
					protoid = group2gid(optarg);
				break;
		}
	}
	if (flags & FL_GRACE && flags & FL_COPY) {
		fputs("Illegal combination of options.\n", stderr);
		usage(slash);
	}
	if ((flags & FL_GRACE && optind != argcnt) || (!(flags & FL_GRACE) && optind >= argcnt)) {
		fputs("Bad number of arguments.\n", stderr);
		usage(slash);
	}
	names = optind;
}

void print_graces(FILE *tf)
{
	char bgbuf[MAXTIMELEN], igbuf[MAXTIMELEN];

	time2str(info.dqi_bgrace, bgbuf);
	time2str(info.dqi_igrace, igbuf);
	fprintf(tf, "Block grace: %s Inode grace: %s\n", bgbuf, igbuf);
}

void print_limits(FILE *tf, qid_t id)
{
	struct dquot dquot;

	read_dquot(id, &dquot);
	fprintf(tf, "Used %luKB, limits: soft=%lu hard=%lu\n", (ulong)dquot.dq_dqb.dqb_curspace,
	  (ulong)dquot.dq_dqb.dqb_bsoftlimit, (ulong)dquot.dq_dqb.dqb_bhardlimit);
	fprintf(tf, "Used %lu inodes, limits: soft=%lu hard=%lu\n", (ulong)dquot.dq_dqb.dqb_curinodes,
	  (ulong)dquot.dq_dqb.dqb_isoftlimit, (ulong)dquot.dq_dqb.dqb_ihardlimit);
}

long parse_date(char *buf, char **res)
{
	char *dot, *errch;
	long tn;

	dot = strchr(buf, ':')+1;
	tn = strtol(dot, &errch, 10);
	if (!(*res = strchr(errch, ' ')))
		goto out;
	if (*errch == ' ' || *errch == 's' || *errch == 'S')
		return tn;
	*((*res)++) = 0;
	if (!strcasecmp(errch, "minutes") || !strcasecmp(errch, "min"))
		return tn*60;
	if (!strcasecmp(errch, "hours") || !strcasecmp(errch, "hour"))
		return tn*3600;
	if (!strcasecmp(errch, "days") || !strcasecmp(errch, "day"))
		return tn*86400;
out:
	die(12, "Can't parse grace time: %s\n", buf);
	return 0;
}

void scan_graces(FILE *tf)
{
	char linebuf[MAXLINELEN], *nextgrace;

	if (!fgets(linebuf, MAXLINELEN, tf))
		die(12, "Unexpected end of file.\n");
	linebuf[strlen(linebuf)-1] = ' ';	/* Overwrite \n */
	info.dqi_bgrace = parse_date(linebuf, &nextgrace);
	info.dqi_igrace = parse_date(nextgrace, &nextgrace);
	mark_quotafile_info_dirty();
}

void scan_limits(FILE *tf, qid_t id)
{
	char linebuf[MAXLINELEN];
	struct dquot dquot;
	ulong usage, softlim, hardlim;

	if (!fgets(linebuf, MAXLINELEN, tf))
		die(12, "Unexpected end of file.\n");
	read_dquot(id, &dquot);
	if (sscanf(linebuf, "Used %luKB, limits: soft=%lu hard=%lu", &usage, &softlim, &hardlim) != 3)
		die(12, "Can't parse limits: %s\n", linebuf);
	dquot.dq_dqb.dqb_bsoftlimit = softlim;
	dquot.dq_dqb.dqb_bhardlimit = hardlim;
	if (!fgets(linebuf, MAXLINELEN, tf))
		die(12, "Unexpected end of file.\n");
	if (sscanf(linebuf, "Used %lu inodes, limits: soft=%lu hard=%lu", &usage, &softlim, &hardlim) != 3)
		die(12, "Can't parse limits: %s\n", linebuf);
	dquot.dq_dqb.dqb_isoftlimit = softlim;
	dquot.dq_dqb.dqb_ihardlimit = hardlim;
	update_times(&dquot.dq_dqb);
	commit_dquot(&dquot, 0);
}

int scan_mountpoint(FILE *tf, char *devname, char *mntpoint)
{
	char linebuf[MAXLINELEN];
	char *ch1, *ch2;

	if (!fgets(linebuf, MAXLINELEN, tf))
		return 1;
	if (!(ch1 = strchr(linebuf, ' ')))
		goto out;
	if (!(ch2 = strchr(++ch1, ' ')))
		goto out;
	*ch2 = 0;
	sstrncpy(devname, ch1, MAXDEVLEN);
	ch2 += 2;
	if (!(ch1 = strchr(ch2, ')')))
		goto out;
	*ch1 = 0;
	sstrncpy(mntpoint, ch2, MAXFILENAMELEN);
	return 0;
out:
	die(12, "Illegal format of edited file. Can't parse: %s\n", linebuf);
	return 0;
}

void copy_quota(int type, qid_t editid)
{
	char devname[MAXDEVLEN];
	char mntpoint[MAXFILENAMELEN];

	init_mount(NULL, 0);
	while (next_mount(type, devname, mntpoint, NULL)) {
		struct dquot dquot, adquot;

		init_io(mntpoint, devname, type);
		read_dquot(protoid, &dquot);
		read_dquot(editid, &adquot);
		adquot.dq_dqb.dqb_bsoftlimit = dquot.dq_dqb.dqb_bsoftlimit;
		adquot.dq_dqb.dqb_bhardlimit = dquot.dq_dqb.dqb_bhardlimit;
		adquot.dq_dqb.dqb_isoftlimit = dquot.dq_dqb.dqb_isoftlimit;
		adquot.dq_dqb.dqb_ihardlimit = dquot.dq_dqb.dqb_ihardlimit;
		update_times(&adquot.dq_dqb);
		commit_dquot(&adquot, 0);
		free_io();
	}
	free_mount();
}

void edit(char *filename)
{
	int pid, stat;

	if ((pid = fork()) < 0)
		die(11, "Canot fork to run editor: %s\n", strerror(errno));
	if (!pid) {
		char *editor;

		if (!(editor = getenv("VISUAL")) && !(editor = getenv("EDITOR")))
			editor = VI_PATH;
		if (execlp(editor, editor, filename, NULL) < 0)
			die(11, "Can't run editor %s: %s\n", editor, strerror(errno));
	}
	waitpid(pid, &stat, 0);
}

void process_name(int type, char *name)
{
	qid_t id = 0;

	if (!(flags & FL_GRACE))
		if (type == QS_USRQUOTA)
			id = name2uid(name);
		else
			id = group2gid(name);

	if (flags & FL_COPY)
		copy_quota(type, id);
	else {
		FILE *tmp;
		int tmpfd;
		char devname[MAXDEVLEN];
		char mntpoint[MAXFILENAMELEN];
		char tmpname[MAXFILENAMELEN];
		char linebuf[MAXLINELEN];

		umask(077);
		sstrncpy(tmpname, _PATH_TMP, MAXFILENAMELEN);
		sstrncat(tmpname, "edqXXXXXX", MAXFILENAMELEN);
		if ((tmpfd = mkstemp(tmpname)) < 0)
			die(10, "Can't create temporary file: %s\n", strerror(errno));
		if (!(tmp = fdopen(tmpfd, "r+")))
			die(10, "Can't reopen temporary file: %s\n", strerror(errno));
		if (flags & FL_GRACE)
			fprintf(tmp, "Edit grace times for %s quota:\n", type2name(type));
		else
			fprintf(tmp, "Edit disk and inode quota for %s %s:\n", type2name(type), name);
		init_mount(NULL, 0);
		while (next_mount(type, devname, mntpoint, NULL)) {
			init_io(mntpoint, devname, type);
			fprintf(tmp, "Device %s (%s):\n", devname, mntpoint);
			if (flags & FL_GRACE)
				print_graces(tmp);
			else
				print_limits(tmp, id);
			free_io();
		}
		free_mount();
		fclose(tmp);
		edit(tmpname);
		if (!(tmp = fopen(tmpname, "r")))
			die(10, "Can't reopen temporary file %s: %s\n", tmpname, strerror(errno));
		fgets(linebuf, MAXLINELEN, tmp);	/* Skip first line */
		while (1) {
			if (scan_mountpoint(tmp, devname, mntpoint))	/* EOF? */
				break;
			init_io(mntpoint, devname, type);
			if (flags & FL_GRACE)
				scan_graces(tmp);
			else
				scan_limits(tmp, id);
			free_io();
		}
		fclose(tmp);
		unlink(tmpname);
	}
}

int main(int argcnt, char **argstr)
{
	parse_options(argcnt, argstr);
	if (flags & FL_GRACE)
		process_name(type, NULL);
	else
		for (;names < argcnt; names++)
			process_name(type, argstr[names]);
	return 0;
}
